# Third Party Stuff
from rest_framework.routers import DefaultRouter

# phone_verify
from .api import VerificationViewSet

default_router = DefaultRouter(trailing_slash=False)
default_router.register("phone", VerificationViewSet, basename="phone")

urlpatterns = list(default_router.urls)



# -*- coding: utf-8 -*-

# Third Party Stuff
from django.apps import apps
from django_dynamic_fixture import G


def create_verification(**kwargs):
    SMSVerification = apps.get_model("phone_verify", "SMSVerification")
    verification = G(SMSVerification, **kwargs)
    return verification



# -*- coding: utf-8 -*-
from __future__ import absolute_import

# Third Party Stuff
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client as TwilioRestClient

# Local
from .base import BaseBackend
from phone_verify.models import SMSVerification


class TwilioBackend(BaseBackend):
    def __init__(self, **options):
        super(TwilioBackend, self).__init__(**options)
        # Lower case it just to be sure
        options = {key.lower(): value for key, value in options.items()}
        self._sid = options.get("sid", None)
        self._secret = options.get("secret", None)  # auth_token
        self._from = options.get("from", None)

        self.client = TwilioRestClient(self._sid, self._secret)
        self.exception_class = TwilioRestException

    def send_sms(self, number, message):
        self.client.messages.create(to=number, body=message, from_=self._from)

    def send_bulk_sms(self, numbers, message):
        for number in numbers:
            self.send_sms(number=number, message=message)


class TwilioSandboxBackend(BaseBackend):
    def __init__(self, **options):
        super(TwilioSandboxBackend, self).__init__(**options)
        # Lower case it just to be sure
        options = {key.lower(): value for key, value in options.items()}
        self._sid = options.get("sid", None)
        self._secret = options.get("secret", None)  # auth_token
        self._from = options.get("from", None)
        self._token = options.get("sandbox_token")

        self.client = TwilioRestClient(self._sid, self._secret)
        self.exception_class = TwilioRestException

    def send_sms(self, number, message):
        self.client.messages.create(to=number, body=message, from_=self._from)

    def send_bulk_sms(self, numbers, message):
        for number in numbers:
            self.send_sms(number=number, message=message)

    def generate_security_code(self):
        """
        Returns a fixed security code
        """
        return self._token

    def validate_security_code(self, security_code, phone_number, session_token):
        return SMSVerification.objects.none(), self.SECURITY_CODE_VALID



# Generated by Django 2.1.7 on 2019-02-14 07:27

import uuid

import phonenumber_field.modelfields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SMSVerification",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("otp", models.CharField(max_length=120, verbose_name="OTP")),
                (
                    "phone_number",
                    phonenumber_field.modelfields.PhoneNumberField(
                        max_length=128, verbose_name="Phone Number"
                    ),
                ),
                (
                    "session_code",
                    models.CharField(
                        max_length=500, verbose_name="Device Session Code"
                    ),
                ),
            ],
            options={
                "verbose_name": "SMS Verification",
                "verbose_name_plural": "SMS Verifications",
                "db_table": "sms_verification",
                "ordering": ("-modified_at",),
            },
        ),
        migrations.AlterUniqueTogether(
            name="smsverification",
            unique_together={("otp", "phone_number", "session_code")},
        ),
    ]



# Standard Library
import uuid

# Third Party Stuff
from django.db import models
from django.utils.translation import ugettext_lazy as _
from phonenumber_field.modelfields import PhoneNumberField


class UUIDModel(models.Model):
    """ An abstract base class model that makes primary key `id` as UUID
    instead of default auto incremented number.
    """

    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)

    class Meta:
        abstract = True


class TimeStampedUUIDModel(UUIDModel):
    """An abstract base class model that provides self-updating
    ``created`` and ``modified`` fields with UUID as primary_key field.
    """

    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    modified_at = models.DateTimeField(auto_now=True, editable=False)

    class Meta:
        abstract = True


class SMSVerification(TimeStampedUUIDModel):
    security_code = models.CharField(_("Security Code"), max_length=120)
    phone_number = PhoneNumberField(_("Phone Number"))
    session_token = models.CharField(_("Device Session Token"), max_length=500)
    is_verified = models.BooleanField(_("Security Code Verified"), default=False)

    class Meta:
        db_table = "sms_verification"
        verbose_name = _("SMS Verification")
        verbose_name_plural = _("SMS Verifications")
        ordering = ("-modified_at",)
        unique_together = ("security_code", "phone_number", "session_token")

    def __str__(self):
        return "{}: {}".format(str(self.phone_number), self.security_code)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

# Third Party Stuff
import nexmo
from nexmo.errors import ClientError

# Local
from .base import BaseBackend
from phone_verify.models import SMSVerification


class NexmoBackend(BaseBackend):
    def __init__(self, **options):
        super().__init__(**options)

        # Lower case it just to be sure
        options = {key.lower(): value for key, value in options.items()}
        self._key = options.get("key", None)
        self._secret = options.get("secret", None)
        self._from = options.get("from", None)

        self.client = nexmo.Client(key=self._key, secret=self._secret)
        self.exception_class = ClientError

    def send_sms(self, number, message):
        self.client.send_message({"from": self._from, "to": number, "text": message})

    def send_bulk_sms(self, numbers, message):
        for number in numbers:
            self.send_sms(number=number, message=message)


class NexmoSandboxBackend(BaseBackend):
    def __init__(self, **options):
        super().__init__(**options)
        # Lower case it just to be sure
        options = {key.lower(): value for key, value in options.items()}
        self._key = options.get("key", None)
        self._secret = options.get("secret", None)
        self._from = options.get("from", None)
        self._token = options.get("sandbox_token", None)

        self.client = nexmo.Client(key=self._key, secret=self._secret)
        self.exception_class = ClientError

    def send_sms(self, number, message):
        self.client.send_message({"from": self._from, "to": number, "text": message})

    def send_bulk_sms(self, numbers, message):
        for number in numbers:
            self.send_sms(number=number, message=message)

    def generate_security_code(self):
        """
        Returns a fixed security code
        """
        return self._token

    def validate_security_code(self, security_code, phone_number, session_token):
        return SMSVerification.objects.none(), self.SECURITY_CODE_VALID



DJANGO_SETTINGS = {
    "SECRET_KEY": "change-me-later",
    "DATABASES": {"default": {"ENGINE": "django.db.backends.sqlite3"}},
    "ROOT_URLCONF": "phone_verify.urls",
    "INSTALLED_APPS": [
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "phone_verify",
    ],
    # PHONE VERIFICATION
    "PHONE_VERIFICATION": {
        "BACKEND": "phone_verify.backends.twilio.TwilioBackend",
        "OPTIONS": {
            "SID": "fake",
            "SECRET": "fake",
            "FROM": "+14755292729",
            "SANDBOX_TOKEN": "123456",
        },
        "TOKEN_LENGTH": 6,
        "MESSAGE": "Welcome to {app}! Please use security code {security_code} to proceed.",
        "APP_NAME": "Phone Verify",
        "SECURITY_CODE_EXPIRATION_TIME": 1,  # In seconds only
        "VERIFY_SECURITY_CODE_ONLY_ONCE": False,
    },
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json
import time

# Third party stuff
import pytest
from django.apps import apps
from django.test import override_settings
from django.urls import reverse

from conftest import backends, sandbox_backends
from . import factories as f
from .test_backends import _get_backend_cls

pytestmark = pytest.mark.django_db

SECURITY_CODE = "123456"
PHONE_NUMBER = "+13478379634"
SESSION_TOKEN = "phone-auth-session-token"


def test_phone_registration_sends_message(client, mocker, backend):
    url = reverse("phone-register")
    phone_number = PHONE_NUMBER
    data = {"phone_number": phone_number}
    twilio_api = mocker.patch(
        "phone_verify.services.PhoneVerificationService.send_verification"
    )

    response = client.post(url, data)

    assert response.status_code == 200
    assert twilio_api.called
    assert "session_token" in response.data
    SMSVerification = apps.get_model("phone_verify", "SMSVerification")
    assert SMSVerification.objects.get(
        session_token=response.data["session_token"], phone_number=phone_number
    )


def test_security_code_session_token_verification_api(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        f.create_verification(
            security_code=SECURITY_CODE,
            phone_number=PHONE_NUMBER,
            session_token=SESSION_TOKEN,
        )
        url = reverse("phone-verify")
        data = {
            "phone_number": PHONE_NUMBER,
            "security_code": SECURITY_CODE,
            "session_token": SESSION_TOKEN,
        }
        response = client.json.post(url, data=data)
        assert response.status_code == 200
        assert response.data["message"] == "Security code is valid."


def test_phone_verification_with_incomplete_payload(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        f.create_verification(
            security_code=SECURITY_CODE,
            phone_number=PHONE_NUMBER,
            session_token=SESSION_TOKEN,
        )
        url = reverse("phone-verify")
        data = {"phone_number": PHONE_NUMBER}
        response = client.json.post(url, data=data)
        assert response.status_code == 400
        response_data = json.loads(json.dumps(response.data))
        assert response_data["session_token"][0] == "This field is required."
        assert response_data["security_code"][0] == "This field is required."

        data = {"security_code": SECURITY_CODE}
        response = client.json.post(url, data=data)
        assert response.status_code == 400
        response_data = json.loads(json.dumps(response.data))
        assert response_data["phone_number"][0] == "This field is required."


def test_phone_verification_with_incorrect_payload(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        f.create_verification(
            security_code=SECURITY_CODE,
            phone_number=PHONE_NUMBER,
            session_token=SESSION_TOKEN,
        )
        url = reverse("phone-verify")
        # Payload with wrong session token
        data = {
            "phone_number": PHONE_NUMBER,
            "security_code": SECURITY_CODE,
            "session_token": "wrong-session-token",
        }
        response = client.json.post(url, data=data)
        response_data = json.loads(json.dumps(response.data))

        backend_cls = _get_backend_cls(backend)

        if backend_cls in backends:
            assert response.status_code == 400
            assert response_data["non_field_errors"][0] == "Session Token mis-match"
        elif backend_cls in sandbox_backends:
            # Sandbox Backend returns a 200 status code as it does not check the payload information
            assert response.status_code == 200

        # Payload with wrong security code
        data = {
            "phone_number": PHONE_NUMBER,
            "security_code": "999999",
            "session_token": SESSION_TOKEN,
        }
        response = client.json.post(url, data=data)
        response_data = json.loads(json.dumps(response.data))
        if backend_cls in backends:
            assert response.status_code == 400
            assert response_data["non_field_errors"][0] == "Security code is not valid"
        elif backend_cls in sandbox_backends:
            # Sandbox Backend returns a 200 status code as it does not check the payload information
            assert response.status_code == 200

        # Payload with incorrect phone_number
        data = {
            "phone_number": "+13478379632",
            "security_code": SECURITY_CODE,
            "session_token": SESSION_TOKEN,
        }
        response = client.json.post(url, data=data)
        response_data = json.loads(json.dumps(response.data))
        if backend_cls in backends:
            assert response.status_code == 400
            assert response_data["non_field_errors"][0] == "Security code is not valid"
        elif backend_cls in sandbox_backends:
            # Sandbox Backend returns a 200 status code as it does not check the payload information
            assert response.status_code == 200


def test_check_security_code_expiry(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        f.create_verification(
            security_code=SECURITY_CODE,
            phone_number=PHONE_NUMBER,
            session_token=SESSION_TOKEN,
        )
        time.sleep(2)
        url = reverse("phone-verify")
        data = {
            "phone_number": PHONE_NUMBER,
            "security_code": SECURITY_CODE,
            "session_token": SESSION_TOKEN,
        }
        response = client.json.post(url, data=data)
        response_data = json.loads(json.dumps(response.data))

        backend_cls = _get_backend_cls(backend)

        if backend_cls in backends:
            assert response.status_code == 400
            assert response_data["non_field_errors"][0] == "Security code has expired"
        elif backend_cls in sandbox_backends:
            # Sandbox Backend returns a 200 status code when verifying security code expiry
            assert response.status_code == 200


def test_verified_security_code(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        f.create_verification(
            security_code=SECURITY_CODE,
            phone_number=PHONE_NUMBER,
            session_token=SESSION_TOKEN,
            is_verified=True,
        )
        url = reverse("phone-verify")
        data = {
            "phone_number": PHONE_NUMBER,
            "security_code": SECURITY_CODE,
            "session_token": SESSION_TOKEN,
        }

        backend_cls = _get_backend_cls(backend)

        # Security code verification is restricted to one time
        backend["VERIFY_SECURITY_CODE_ONLY_ONCE"] = True
        response = client.json.post(url, data=data)
        response_data = json.loads(json.dumps(response.data))
        if backend_cls in backends:
            assert response.status_code == 400
            assert (
                response_data["non_field_errors"][0]
                == "Security code is already verified"
            )
        elif backend_cls in sandbox_backends:
            # Sandbox Backend returns a 200 status code when verifying security code
            assert response.status_code == 200

        # Security code verification is not restricted to one time
        backend["VERIFY_SECURITY_CODE_ONLY_ONCE"] = False
        response = client.json.post(url, data=data)
        response_data = json.loads(json.dumps(response.data))
        assert response.status_code == 200
        assert response.data["message"] == "Security code is valid."



# Generated by Django 2.1.11 on 2019-08-17 17:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("phone_verify", "0001_initial")]

    operations = [
        migrations.RenameField(
            model_name="SMSVerification", old_name="otp", new_name="security_code"
        ),
        migrations.RenameField(
            model_name="SMSVerification",
            old_name="session_code",
            new_name="session_token",
        ),
        migrations.AlterField(
            model_name="SMSVerification",
            name="security_code",
            field=models.CharField(max_length=120, verbose_name="Security Code"),
        ),
        migrations.AddField(
            model_name="smsverification",
            name="is_verified",
            field=models.BooleanField(
                default=False, verbose_name="Security Code Verified"
            ),
        ),
        migrations.AlterField(
            model_name="smsverification",
            name="session_token",
            field=models.CharField(max_length=500, verbose_name="Device Session Token"),
        ),
    ]



# -*- coding: utf-8 -*-

import importlib

# Third Party Stuff
import pytest
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings
from nexmo.errors import ClientError
from twilio.base.exceptions import TwilioRestException

# phone_verify Stuff
import phone_verify.services
from phone_verify.services import (
    PhoneVerificationService,
    send_security_code_and_generate_session_token,
)

from .test_backends import _get_backend_cls

pytestmark = pytest.mark.django_db


def test_message_generation_and_sending_service(client, mocker, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        service = PhoneVerificationService(phone_number="+13478379634")
        backend_service = backend.get("BACKEND")
        mock_api = mocker.patch(f"{backend_service}.send_sms")
        service.send_verification("+13478379634", "123456")

        assert mock_api.called


def test_exception_is_logged_when_raised(client, mocker, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        mock_send_verification = mocker.patch(
            "phone_verify.services.PhoneVerificationService.send_verification"
        )
        mock_logger = mocker.patch("phone_verify.services.logger")
        backend_cls = _get_backend_cls(backend)
        if (
            backend_cls == "nexmo.NexmoBackend"
            or backend_cls == "nexmo.NexmoSandboxBackend"
        ):
            exc = ClientError()
            mock_send_verification.side_effect = exc
        elif (
            backend_cls == "twilio.TwilioBackend"
            or backend_cls == "twilio.TwilioSandboxBackend"
        ):
            exc = TwilioRestException(status=mocker.Mock(), uri=mocker.Mock())
            mock_send_verification.side_effect = exc
        send_security_code_and_generate_session_token(phone_number="+13478379634")
        mock_logger.error.assert_called_once_with(
            f"Error in sending verification code to +13478379634: {exc}"
        )


@override_settings(
    PHONE_VERIFICATION={
        "BACKEND": "phone_verify.backends.twilio.TwilioBackend",
        "MESSAGE": "Welcome to {app}! Please use security code {security_code} to proceed.",
        "APP_NAME": "Phone Verify",
        "SECURITY_CODE_EXPIRATION_TIME": 1,  # In seconds only
        "VERIFY_SECURITY_CODE_ONLY_ONCE": False,
    }
)
def test_exception_is_raised_when_improper_settings(client):
    with pytest.raises(ImproperlyConfigured) as exc:
        PhoneVerificationService(phone_number="+13478379634")
        assert (
            exc.info
            == "Please specify following settings in settings.py: OPTIONS, TOKEN_LENGTH"
        )


def test_exception_is_raised_when_no_settings(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        del settings.PHONE_VERIFICATION
        with pytest.raises(ImproperlyConfigured) as exc:
            importlib.reload(phone_verify.services)
            PhoneVerificationService(phone_number="+13478379634")
            assert exc.info == "Please define PHONE_VERIFICATION in settings"



# -*- coding: utf-8 -*-

import random
from abc import ABCMeta, abstractmethod

# Third Party Stuff
import jwt
from django.conf import settings as django_settings
from django.utils import timezone
from django.utils.crypto import get_random_string

from ..models import SMSVerification

DEFAULT_TOKEN_LENGTH = 6


class BaseBackend(metaclass=ABCMeta):
    SECURITY_CODE_VALID = 0
    SECURITY_CODE_INVALID = 1
    SECURITY_CODE_EXPIRED = 2
    SECURITY_CODE_VERIFIED = 3
    SESSION_TOKEN_INVALID = 4

    def __init__(self, **settings):
        self.exception_class = None

    @abstractmethod
    def send_sms(self, numbers, message):
        raise NotImplementedError()

    @abstractmethod
    def send_bulk_sms(self, numbers, message):
        raise NotImplementedError()

    @classmethod
    def generate_security_code(cls):
        """
        Returns a unique random `security_code` for given `TOKEN_LENGTH` in the settings.
        """
        token_length = django_settings.PHONE_VERIFICATION.get(
            "TOKEN_LENGTH", DEFAULT_TOKEN_LENGTH
        )
        return get_random_string(token_length, allowed_chars="0123456789")

    @classmethod
    def generate_session_token(cls, phone_number):
        """
        Returns a unique session_token for
        identifying a particular device in subsequent calls.
        """
        data = {"phone_number": phone_number, "nonce": random.random()}
        return jwt.encode(data, django_settings.SECRET_KEY).decode()

    @classmethod
    def check_security_code_expiry(cls, stored_verification):
        """
        Returns True if the `security_code` for the `stored_verification` is expired.
        """
        time_difference = timezone.now() - stored_verification.created_at
        if time_difference.seconds > django_settings.PHONE_VERIFICATION.get(
            "SECURITY_CODE_EXPIRATION_TIME"
        ):
            return True
        return False

    def create_security_code_and_session_token(self, number):
        """
        Creates a temporary `security_code` and `session_token` inside the DB.

        `security_code` is the code that user would enter to verify their phone_number.
        `session_token` is used to verify if the subsequent call for verification is
        by the same device that initiated a phone number verification in the
        first place.

        :param number: Phone number of recipient

        :return security_code: string of sha security_code
        :return session_token: string of session_token
        """
        security_code = self.generate_security_code()
        session_token = self.generate_session_token(number)

        # Delete old security_code(s) for phone_number if already exists
        SMSVerification.objects.filter(phone_number=number).delete()

        # Default security_code generated of 6 digits
        SMSVerification.objects.create(
            phone_number=number,
            security_code=security_code,
            session_token=session_token,
        )
        return security_code, session_token

    def validate_security_code(self, security_code, phone_number, session_token):
        """
        A utility method to verify if the `security_code` entered is valid for
        a given `phone_number` along with the `session_token` used.

        :param security_code: Security code entered for verification
        :param phone_number: Phone number to be verified
        :param session_token: Session token to identify the device

        :return stored_verification: Contains the verification object
        corresponding to the phone_number if found, else None.
        :return status: Status for the stored_verification object.
        Can be one of the following:
            - `BaseBackend.SECURITY_CODE_VALID`
            - `BaseBackend.SECURITY_CODE_INVALID`
            - `BaseBackend.SECURITY_CODE_EXPIRED`
            - `BaseBackend.SECURITY_CODE_VERIFIED`
            - `BaseBackend.SESSION_TOKEN_INVALID`
        """
        stored_verification = SMSVerification.objects.filter(
            security_code=security_code, phone_number=phone_number
        ).first()

        # check security_code exists
        if stored_verification is None:
            return stored_verification, self.SECURITY_CODE_INVALID

        # check session code exists
        if not stored_verification.session_token == session_token:
            return stored_verification, self.SESSION_TOKEN_INVALID

        # check security_code is not expired
        if self.check_security_code_expiry(stored_verification):
            return stored_verification, self.SECURITY_CODE_EXPIRED

        # check security_code is not verified
        if stored_verification.is_verified and django_settings.PHONE_VERIFICATION.get(
            "VERIFY_SECURITY_CODE_ONLY_ONCE"
        ):
            return stored_verification, self.SECURITY_CODE_VERIFIED

        # mark security_code as verified
        stored_verification.is_verified = True
        stored_verification.save()

        return stored_verification, self.SECURITY_CODE_VALID



import pytest
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings
from django.urls import reverse
from django.utils.module_loading import import_string

from conftest import sandbox_backends
from phone_verify.backends.base import BaseBackend

PHONE_NUMBER = "+13478379634"
SECURITY_CODE = "123456"
SESSION_TOKEN = "phone-auth-session-token"

pytestmark = pytest.mark.django_db


def _get_backend_cls(backend):
    backend_import_path = backend.get("BACKEND")
    backend_cls = backend_import_path.split("phone_verify.backends.")[1]
    return backend_cls


def test_backends(client, mocker, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        url = reverse("phone-register")
        phone_number = PHONE_NUMBER
        data = {"phone_number": phone_number}

        mocker.patch(
            "phone_verify.backends.base.BaseBackend.generate_session_token",
            return_value=SESSION_TOKEN,
        )
        mocker.patch(
            "phone_verify.backends.base.BaseBackend.generate_security_code",
            return_value=SECURITY_CODE,
        )
        message = "Welcome to Phone Verify! Please use security code 123456 to proceed."
        from_number = settings.PHONE_VERIFICATION["OPTIONS"]["FROM"]

        backend_cls = _get_backend_cls(backend)

        if (
            backend_cls == "nexmo.NexmoBackend"
            or backend_cls == "nexmo.NexmoSandboxBackend"
        ):
            # Mock the nexmo client
            mock_nexmo_send_message = mocker.patch(
                "phone_verify.backends.nexmo.nexmo.Client.send_message"
            )
            test_data = {"from": from_number, "to": phone_number, "text": message}
        elif (
            backend_cls == "twilio.TwilioBackend"
            or backend_cls == "twilio.TwilioSandboxBackend"
        ):
            # Mock the twilio client
            mock_twilio_send_message = mocker.patch(
                "phone_verify.backends.twilio.TwilioRestClient.messages"
            )
            mock_twilio_send_message.create = mocker.MagicMock()

        response = client.post(url, data)
        assert response.status_code == 200

        if backend_cls == "nexmo.NexmoBackend":
            mock_nexmo_send_message.assert_called_once_with(test_data)
        elif backend_cls == "twilio.TwilioBackend":
            mock_twilio_send_message.create.assert_called_once_with(
                to=phone_number, body=message, from_=from_number
            )

        # Get the last part of the backend and check if that is a Sandbox Backend
        if backend_cls in sandbox_backends:
            url = reverse("phone-verify")
            data = {
                "phone_number": phone_number,
                "session_token": SESSION_TOKEN,
                "security_code": SECURITY_CODE,
            }

            response = client.post(url, data)
            assert response.status_code == 200
            response_data = {"message": "Security code is valid."}
            assert response.data == response_data


def test_error_raised_when_no_backend_specified(client, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        settings.PHONE_VERIFICATION["BACKEND"] = None
        url = reverse("phone-register")
        phone_number = PHONE_NUMBER
        data = {"phone_number": phone_number}
        with pytest.raises(ImproperlyConfigured) as exc:
            client.post(url, data)
            assert (
                exc.info
                == "Please specify BACKEND in PHONE_VERIFICATION within your settings"
            )


def test_send_bulk_sms(client, mocker, backend):
    with override_settings(PHONE_VERIFICATION=backend):
        backend_import = settings.PHONE_VERIFICATION["BACKEND"]
        backend_cls = import_string(backend_import)
        cls_obj = backend_cls(**settings.PHONE_VERIFICATION["OPTIONS"])

        mock_send_sms = mocker.patch(f"{backend_import}.send_sms")
        numbers = ["+13478379634", "+13478379633", "+13478379632"]
        message = "Fake message"

        cls_obj.send_bulk_sms(numbers, message)
        assert mock_send_sms.called
        assert mock_send_sms.call_count == 3
        mock_send_sms.assert_has_calls(
            [
                mocker.call(number=numbers[0], message=message),
                mocker.call(number=numbers[1], message=message),
                mocker.call(number=numbers[2], message=message),
            ]
        )


class TestBaseBackend(BaseBackend):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


def test_base_backend_abstract_methods(mocker):
    abstract_methods = BaseBackend.__abstractmethods__
    TestBaseBackend.__abstractmethods__ = frozenset()
    cls_obj = TestBaseBackend()

    for method in abstract_methods:
        with pytest.raises(NotImplementedError):
            # This might fail in case, we have methods having different
            # number of arguments
            getattr(cls_obj, method)(mocker.Mock(), mocker.Mock())



# -*- coding: utf-8 -*-

# Third Party Stuff
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.permissions import AllowAny

from .base import response
from .serializers import PhoneSerializer, SMSVerificationSerializer
from .services import send_security_code_and_generate_session_token


class VerificationViewSet(viewsets.GenericViewSet):
    @action(
        detail=False,
        methods=["POST"],
        permission_classes=[AllowAny],
        serializer_class=PhoneSerializer,
    )
    def register(self, request):
        serializer = PhoneSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        session_token = send_security_code_and_generate_session_token(
            str(serializer.validated_data["phone_number"])
        )
        return response.Ok({"session_token": session_token})

    @action(
        detail=False,
        methods=["POST"],
        permission_classes=[AllowAny],
        serializer_class=SMSVerificationSerializer,
    )
    def verify(self, request):
        serializer = SMSVerificationSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        return response.Ok({"message": "Security code is valid."})



# -*- coding: utf-8 -*-

# Standard Library
import logging

# Third Party Stuff
from django.utils.translation import ugettext as _
from phonenumber_field.serializerfields import PhoneNumberField
from rest_framework import serializers

# Phone Auth Stuff
from .backends import get_sms_backend

logger = logging.getLogger(__name__)


class PhoneSerializer(serializers.Serializer):
    phone_number = PhoneNumberField()


class SMSVerificationSerializer(serializers.Serializer):
    phone_number = PhoneNumberField(required=True)
    session_token = serializers.CharField(required=True)
    security_code = serializers.CharField(required=True)

    def validate(self, attrs):
        attrs = super().validate(attrs)
        phone_number = attrs.get("phone_number", None)
        security_code, session_token = (
            attrs.get("security_code", None),
            attrs.get("session_token", None),
        )
        backend = get_sms_backend(phone_number=phone_number)
        verification, token_validatation = backend.validate_security_code(
            security_code=security_code,
            phone_number=phone_number,
            session_token=session_token,
        )

        if verification is None:
            raise serializers.ValidationError(_("Security code is not valid"))
        elif token_validatation == backend.SESSION_TOKEN_INVALID:
            raise serializers.ValidationError(_("Session Token mis-match"))
        elif token_validatation == backend.SECURITY_CODE_EXPIRED:
            raise serializers.ValidationError(_("Security code has expired"))
        elif token_validatation == backend.SECURITY_CODE_VERIFIED:
            raise serializers.ValidationError(_("Security code is already verified"))

        return attrs



# Third Party Stuff
import rest_framework.response


class Response(rest_framework.response.Response):
    """The various HTTP responses for use in returning proper HTTP codes.
    """

    def __init__(
        self,
        data=None,
        status=None,
        template_name=None,
        headers=None,
        exception=False,
        content_type=None,
    ):
        super().__init__(data, status, template_name, headers, exception, content_type)


class Ok(Response):
    """200 OK

    Should be used to indicate nonspecific success. Must not be used to
    communicate errors in the response body.
    In most cases, 200 is the code the client hopes to see. It indicates that
    the REST API successfully carried out whatever action the client requested,
    and that no more specific code in the 2xx series is appropriate. Unlike
    the 204 status code, a 200 response should include a response body.
    """

    status_code = 200



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django-phone-verify",
    version="2.0.0",
    packages=find_packages(),
    include_package_data=True,
    license="GPLv3",
    description="A Django app to support phone number verification using security code sent via SMS.",
    long_description=README,
    url="https://github.com/CuriousLearner/django-phone-verify",
    author="Sanyam Khurana",
    author_email="sanyam@sanyamkhurana.com",
    install_requires=[
        "django>=2.1.5",
        "djangorestframework>=3.9.0",
        "PyJWT>=1.7.1",
        "python-dotenv>=0.10.0",
        "phonenumbers>=8.10.2",
        "django-phonenumber-field>=2.1.0",
        "twilio>=6.21.0",
        "nexmo>=2.4.0",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 2.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

# Third Party Stuff
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

# phone_verify stuff
from .backends import get_sms_backend

logger = logging.getLogger(__name__)

# iOS uses "session code" to parse the security code and support copying on clipboard.
DEFAULT_MESSAGE = (
    "Welcome to {app}, use session code {security_code} for authentication."
)
DEFAULT_APP_NAME = "Phone Verify"


class PhoneVerificationService(object):
    try:
        phone_settings = settings.PHONE_VERIFICATION
    except AttributeError:
        raise ImproperlyConfigured("Please define PHONE_VERIFICATION in settings")

    verification_message = phone_settings.get("MESSAGE", DEFAULT_MESSAGE)

    def __init__(self, phone_number, backend=None):
        self._check_required_settings()
        if backend is None:
            self.backend = get_sms_backend(phone_number=phone_number)

    def send_verification(self, number, security_code):
        """
        Send a verification text to the given number to verify.

        :param number: the phone number of recipient.
        """
        message = self._generate_message(security_code)

        self.backend.send_sms(number, message)

    def _generate_message(self, security_code):
        return self.verification_message.format(
            app=settings.PHONE_VERIFICATION.get("APP_NAME", DEFAULT_APP_NAME),
            security_code=security_code,
        )

    def _check_required_settings(self):
        required_settings = {
            "BACKEND",
            "OPTIONS",
            "TOKEN_LENGTH",
            "MESSAGE",
            "APP_NAME",
            "SECURITY_CODE_EXPIRATION_TIME",
            "VERIFY_SECURITY_CODE_ONLY_ONCE",
        }
        user_settings = set(settings.PHONE_VERIFICATION.keys())
        if not required_settings.issubset(user_settings):
            raise ImproperlyConfigured(
                "Please specify following settings in settings.py: {}".format(
                    ", ".join(required_settings - user_settings)
                )
            )


def send_security_code_and_generate_session_token(phone_number):
    sms_backend = get_sms_backend(phone_number)
    security_code, session_token = sms_backend.create_security_code_and_session_token(
        phone_number
    )
    service = PhoneVerificationService(phone_number=phone_number)
    try:
        service.send_verification(phone_number, security_code)
    except service.backend.exception_class as exc:
        logger.error(
            "Error in sending verification code to {phone_number}: "
            "{error}".format(phone_number=phone_number, error=exc)
        )
    return session_token



import pytest

from tests import factories as f

pytestmark = pytest.mark.django_db

SECURITY_CODE = "123456"
PHONE_NUMBER = "+13478379634"
SESSION_TOKEN = "phone-auth-session-token"


def test_create_sms_verification(client, mocker, backend):
    sms_verification = f.create_verification(
        security_code=SECURITY_CODE,
        phone_number=PHONE_NUMBER,
        session_token=SESSION_TOKEN,
    )
    assert str(sms_verification) == f"{PHONE_NUMBER}: {SECURITY_CODE}"
